module("smp_aes", package.seeall)

_VERSION = '0.1'

local ffi = require "ffi"
local ffi_new = ffi.new
local ffi_str = ffi.string
local C = ffi.C

local mt = { __index = smp_aes }

ffi.cdef[[
  int sprintf(char *buf, const char *format, ...);
  int smp_aes_decode(const unsigned char *input, int size, unsigned char *output);
  int smp_aes_encode(const unsigned char *input, int size, unsigned char *output);
]]
local ffi_str_buf = ffi_new("uint8_t[?]", 1024)
local smp_aes_lib = ffi.load("/soooner/nginx/lualib/self/libsmp_aes.so")

function new(self)
    return setmetatable({ _ctx = ctx }, mt)
end

function encode(self, data)
  local len = smp_aes_lib.smp_aes_encode(ffi.cast("char *", data), #data, ffi_str_buf)
  return ffi.string(ffi_str_buf, len)
end

function decode(self, data)
  local len = smp_aes_lib.smp_aes_decode(data, #data, ffi_str_buf)
  if len > 0 then
    return ffi.string(ffi_str_buf, len)
  end
  return nil
end

getmetatable(smp_aes).__newindex = function (table, key, val)
    error('attempt to write to undeclared variable "' .. key .. '": '
            .. debug.traceback())
end
